import numpy as np
import warnings
from subprocess import Popen, PIPE
import os
import re

__flag__ = None
__tshark_min_version__ = '2.6.0'
__tshark_current_version__  = ''
__numpy_min_version__ = '1.18.0'
class Reader(object):
    """Reader object for extracting features from .pcap files

        Attributes
        ----------
        verbose : boolean
            Boolean indicating whether to be verbose in reading
    """

    ########################################################################
    #                         Class initialisation                         #
    ########################################################################

    def __init__(self, verbose=False):
        """Reader object for extracting features from .pcap files

            Parameters
            ----------
            verbose : boolean, default=False
                Boolean indicating whether to be verbose in reading
            """
        # Set verbosity level
        self.verbose = verbose

    ########################################################################
    #                             Read method                              #
    ########################################################################

    def read(self, path,filter="",extension="",ip_layer =False):
        """Read TCP and UDP packets from .pcap file given by path.
            Automatically choses fastest available backend to use.

            Parameters
            ----------
            path : string
                Path to .pcap file to read.

            filter : string
                filter condition to be passed to tshark

            extension : string or list (of string)
                Additional field(s) to be extracted, besides the default fields.
                The field name is consistent with that of Wireshark, such as tls.handshake.extension_server_name means the SNI of TLS flow.
                If type(extension) is string, then only one extra field will be extracted.
                If type(extension) is list of string, then multi fileds will be extracted.

            Returns
            -------
            result : np.array of shape=(n_packets, n_features)
                Where features consist of:

                0) Filename of capture
                1) Protocol TCP/UDP
                2) TCP/UDP stream identifier
                3) Timestamp of packet
                4) Length of packet
                5) IP packet source
                6) IP packet destination
                7) TCP/UDP packet source port
                8) TCP/UDP packet destination port
                9) TCP length
                10) UDP length
                11) extension(s)

            Warning
            -------
            warning
                Method throws warning if tshark is not available.
            """

        # If verbose, print which file is currently being read
        if self.verbose:
            print("Reading {}...".format(path))

        # Check if we can use fast tshark read or slow pyshark read
        try:
            if  os.path.exists(path) == False:
                raise FileExistsError('file {0} does not exist.'.format(path))

            if __flag__ == None:
                # Call Tshark on packets
                command = 'tshark -v'
                try:
                    process = Popen(command, stdout=PIPE, stderr=PIPE)
                    # Get output
                    out, err = process.communicate()
                except :
                    raise  EnvironmentError('tshark is not installed or added to environment path.')
                head = out.decode("utf-8").split('\n')[0].strip()
                version = re.findall('TShark \(Wireshark\) (.*?) \(',head,re.DOTALL)[0]
                if version < __tshark_min_version__ :
                    raise  EnvironmentError('the version of tshark (wireshark) should be greater than {1} at least, however the current version is {0}.'.format(version,__tshark_min_version__))
                __tshark_current_version__ = version
                if np.__version__ < __numpy_min_version__ :
                    raise  EnvironmentError('the version of numpy should be greater than {1} at least, however the current version is {0}.'.format(np.__version__ , __numpy_min_version__))

                __flag__ == object()
            return self.read_tshark(path,filter,extension,ip_layer)
        except Exception as ex:
            if isinstance(ex,EnvironmentError):
                raise EnvironmentError(ex)
            warnings.warn("Running Error : tshark parse error : '{0}'."
                          .format(ex))

    def read_tshark(self, path,filter_str="",extension="",ip_layer =False):
        """Read TCP and UDP packets from file given by path using tshark backend

            Parameters
            ----------
            path : string
                Path to .pcap file to read.

            Returns
            -------
            result : np.array of shape=(n_packets, n_features)
                Where features consist of:

                0) Filename of capture
                1) Protocol TCP/UDP
                2) TCP/UDP stream identifier
                3) Timestamp of packet
                4) Length of packet
                5) IP packet source
                6) IP packet destination
                7) TCP/UDP packet source port
                8) TCP/UDP packet destination port
                9) SSL/TLS certificate if exists, else None
                10) TCP length
                11) UDP length
            """
        # Create Tshark command
        if ip_layer == False:
            command = ["tshark", "-r", path, "-Tfields", "-E", "separator=+",
                   "-e", "frame.time_epoch",
                   "-e", "tcp.stream",
                   "-e", "udp.stream", #only output one line
                   "-e", "ip.proto",
                   "-e", "ip.src",
                   "-e", "tcp.srcport",
                   "-e", "udp.srcport", #only output one line
                   "-e", "ip.dst",
                   "-e", "tcp.dstport",
                   "-e", "udp.dstport", #only output one line
                   "-e", "ip.len",
                   '-e', "tcp.len",
                   "-e", "udp.length",   #only output one line,
                   "-e", 'ip.id',
                   "-2","-R", "ip and not icmp and  not tcp.analysis.retransmission and not tcp.analysis.out_of_order and not tcp.analysis.duplicate_ack and not mdns and not ssdp{0}"]
        else:
            command = ["tshark", "-r", path, "-Tfields", "-E", "separator=+",
                   "-e", "frame.time_epoch",
                   "-e", "tcp.stream",
                   "-e", "udp.stream", #only output one line
                   "-e", "ip.proto",
                   "-e", "ip.src",
                   "-e", "tcp.srcport",
                   "-e", "udp.srcport", #only output one line
                   "-e", "ip.dst",
                   "-e", "tcp.dstport",
                   "-e", "udp.dstport", #only output one line
                   "-e", "ip.len",
                   '-e', "tcp.len",
                   "-e", "udp.length",   #only output one line,
                   "-e", 'ip.id',
                   "-2","-R", "ip and not icmp{0}"]

        if filter_str != "":
            command[-1] = command[-1].format(" and "+filter_str)
        else:
            command[-1] = command[-1].format("")
        #Add extension fields
        if type(extension) == type(""):
            extension  = [extension]

        for each in extension:
            if each!="":
                command.insert(-5,'-e')
                command.insert(-5,each)
        #print(" ".join(command))
        # Initialise result


        result = list()

        # Call Tshark on packets
        process = Popen(command, stdout=PIPE, stderr=PIPE)
        # Get output
        out, err = process.communicate()

        # Give warning message if any
        if err:
            warnings.warn("Error reading file: '{}'".format(
                err.decode('utf-8')))
        protocols = {'17': 'udp', '6': 'tcp','47':'gre'}
        # Read each packet
        for packet in filter(None, out.decode('utf-8').split('\n')):
            # Get all data from packets
            packet = packet.strip()
            packet = packet.split('+')
            #print(packet)
            if len(packet) < 14: continue

            # Perform check on multiple ip addresses
            packet[3] = protocols.get(packet[3],'unknown')
            packet[4] = packet[4].split(',')[0]             #ip.src
            packet[7] = packet[7].split(',')[0]             #ip.dst
            packet[10] = packet[10].replace(',', '')        #ip.len
            #if packet[2]=='udp':
            #    print('#' * 10)
            #    print(packet)

            # Add packet to result
            #路径|tcp(udp)|flowid|时间戳|IP长度|srcIP|dstIP|srcport|dstport|payload长度|extension|
            if packet[3]=='tcp' and ip_layer == False:
                result.append([path]+[packet[3],packet[1],packet[0],packet[10],packet[4],packet[7],packet[5],packet[8],packet[11],packet[13:-1]])
            elif packet[3]=='udp' and ip_layer == False:
                result.append([path] +[packet[3],packet[2],packet[0],packet[10],packet[4],packet[7],packet[6],packet[9],packet[12],packet[13:-1]])
            else:   #非tcp,udp协议,那就只提取ip-len。对于此类flowid填0,源端口填1,目的端口填0
                result.append([path] +[packet[3],0,packet[0],packet[10],packet[4],packet[7],1,0,packet[10],packet[13:-1]])
            #print(result[-1])

        # Get result as numpy array

        result = np.asarray(result)

        # Check if any items exist
        if not result.shape[0]:
            return np.zeros((0, 12+len(extension)), dtype=object)

        # Change protocol number to text

        #print(result.shape)
        #print(result[0:2, [0, 3, 2, 1, 8, 4, 6, 5, 7, 9,10]])
        # Return in original order

        return result



#coding:utf8
__author__ = 'dk'
import setuptools
long_desp = \
'''
A python lib to parse traffic flow information from pcaps.\n
Homepage : https://github.com/jmhIcoding/flowcontainer.\n
Fix bugs:\n
\t set the default filter string to be `tcp or udp or gre`.\n
\t update help information for errors. \n
'''

setuptools.setup(
    name="flowcontainer",
    version="3.7",
    author="Minghao Jiang",
    author_email="jiangminghao@iie.ac.cn",
    description="A python lib to parse traffic flow information from pcaps",
    url="https://github.com/jmhIcoding/flowcontainer",
    long_description=long_desp,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



try:
    from .flows import Flow
except:
    try:
        from flows import Flow
    except Exception as e:
        raise ValueError(e)

class FlowGenerator(object):
    """Generator for Flows from packets extraced using reader.Reader.read()"""

    def combine(self, packets,extention=[]):
        """Combine individual packets into a flow representation

            Parameters
            ----------
            packets : np.array of shape=(n_samples_packets, n_features_packets)
                Output from Reader.read

            Returns
            -------
            flows : dict
                Dictionary of flow_key -> Flow()
            """
        # Initialise result
        result = dict()

        # For each packet, add it to a flow
        for packet in packets:
            key = (packet[0], packet[1], packet[2])
            # Add packet to flow
            result[key] = result.get(key, Flow()).add(packet,extention)
        #Remove empty payload flow
        insuitable_flow = list()
        for each in result:
            if len(result[each].payload_lengths) <= 0 :
                insuitable_flow.append(each)
        for each in insuitable_flow:
            result.pop(each)
        # Return result
        return result



from collections import Counter

class NetworkDestination(object):
    """NetworkDestination object for flow samples

        Attributes
        ----------
        identifier : object
            Unique identifier for NetworkDestination

        samples : list
            List of flows stored in NetworkDestination

        destinations : set
            Set of destination (IP, port) tuples related to NetworkDestination

        certificates : set
            Set of TLS certificates related to NetworkDestination

        labels : Counter
            Labels related to NetworkDestination
    """

    def __init__(self, identifier, samples=[]):
        """NetworkDestination object for flow samples

            Parameters
            ----------
            identifier : object
                Identifier for NetworkDestination
                Important: identifier must be unique!

            samples : iterable of Flow
                Samples to store in this NetworkDestination.
            """
        # Initialise variables
        self.identifier   = identifier
        self.samples      = []
        self.destinations = set()
        self.certificates = set()
        self.labels       = Counter()

        # Add each datapoint
        for X in samples:
            self.add(X)

    ########################################################################
    #                         Add flows to cluster                         #
    ########################################################################

    def add(self, X, y=None):
        """Add flow X to NetworkDestination object.

            Parameters
            ----------
            X : Flow
                Datapoint to store in this NetworkDestination.

            y : object
                Label for datapoint
            """
        # Add datapoint
        self.samples.append(X)
        self.labels.update([y])
        # Update pointers
        self.destinations.add(X.destination)
        self.certificates.add(X.certificate)


    def merge(self, other):
        """Merge NetworkDestination with other NetworkDestination object.

            Parameters
            ----------
            other : NetworkDestination
                Other NetworkDestination object to merge with.
            """
        # Only merge in case other is NetworkDestination object
        if isinstance(other, NetworkDestination):
            # Merge two NetworkDestinations
            self.samples.extend(other.samples)
            # Merge pointers
            self.destinations |= other.destinations
            self.certificates |= other.certificates
            self.labels += other.labels

    ########################################################################
    #                           Get description                            #
    ########################################################################

    def get_description(self):
        """Returns human readable description of cluster"""
        # Get descriptions
        descr_cert = [X.certificate for X in self.samples]
        descr_ip   = ["{}".format(X.destination) for X in self.samples]
        # Remove None values
        descr_cert = [x for x in descr_cert if x is not None]
        descr_ip   = [x for x in descr_ip   if x is not None]
        # Get most common
        descr_cert = Counter(descr_cert).most_common(1)
        descr_ip   = Counter(descr_ip  ).most_common(1)
        # Return description
        try   : return descr_cert[0][0]
        except: return descr_ip[0][0]

    ########################################################################
    #                           Object overrides                           #
    ########################################################################

    def __str__(self):
        """Returns string presentation of self."""
        return "NetworkDestination [{:4}] [size={:4}] [IPs={}] [labels={}]".\
                format(self.identifier, len(self.samples),
                list(sorted(self.destinations)), self.labels)



__author__ = 'dk'
from flowcontainer.reader import Reader
from flowcontainer.flow_generator import FlowGenerator
def extract(infile,filter="(tcp or udp or gre)",extension="", ip_layer=False):
    """Extract flows from given pcap file.

        Parameters
        ----------
        infile : string
            Path to input file.
        filter : string
            Filter condition, which is the same with wireshark
        extension : string or (list of string)
            Additional field(s) to be extracted, besides the default fields.
            The field name is consistent with that of Wireshark, such as tls.handshake.extension_server_name means the SNI of TLS flow.
            If type(extension) is string, then only one extra field will be extracted.
            If type(extension) is list of string, then multi fileds will be extracted.

        Returns
        -------
        result : dict
            Dictionary of flow_key -> flow.
        """
    if type(extension)==type(""):
        extension = [extension]

    for each in extension:
        if type(each)!= type(""):
            raise TypeError("extension must be string!")

    reader = Reader(verbose=True)
    flow_generator = FlowGenerator()
    # Read packets
    result = reader.read(infile,filter,extension,ip_layer)
    # Combine packets into flows
    result = flow_generator.combine(result,extension)
    # Return result
    return result


from datetime import datetime
import ipaddress

################################################################################
#                              Single Flow object                              #
################################################################################

class Flow(object):
    """Flow object extracted from pcap file that can be used for fingerprinting

        Attributes
        ----------
        src : string
            Source IP

        sport : int
            Source port

        dst : string
            Destination IP

        dport : int
            Destination port

        source : tuple
            (Source IP, source port) tuple

        destination : tuple
            (Destination IP, destination port) tuple

        time_start : int
            Timestamp of first packet in flow

        time_end : int
            Timestamp of last packet in flow

        certificate : Object
            Certificate of flow, if any

        ip_lengths : list
            List of packet length for each ip packet in flow
        payload_lengths : list
            List of payload sequence for each tcp/udp fragment with non-zero payload in flow.
        ip_timestamps : list
            List of timestamps corresponding to each ip packet in flow, it may contain packets without any tcp/udp payload.

        payload_timestamps: list
            List of timestamps corresponding to each tcp/udp fragment with non-zero payload in flow.
    """

    def __init__(self,main='payload'):
        """
            param
            -----------
            main: str
              'payload' means the main lengths sequence and timestampes sequence refer to packets with non-zero payload, the sequences will fitler out zero payload packets.
              'ip'   means the main lengths sequence and timestamps sequence refer to any packets, it will not filter any packets.
        """
        self.main = main
        """Initialise an empty Flow."""
        # Initialise flow endpoints
        self.src   = None
        self.sport = None
        self.dst   = None
        self.dport = None

        # Initialise extension
        self.extension = dict()
        # Initialise packet lengths
        self.ip_lengths   = list()
        self.payload_lengths = list()
        # Initialise packet timestamps
        self.ip_timestamps = list()
        self.payload_timestamps = list()   #non-zero payload packet's timestamp sequence.

        # Refer the main property
        self.lengths = self.payload_lengths if main=='payload' else self.ip_lengths
        self.timestamps = self.payload_timestamps if main=='payload' else self.ip
    ########################################################################
    #                        Add new packet to flow                        #
    ########################################################################

    def add(self, packet,extension):
        """Add a new packet to the flow.

            Parameters
            ----------
            packet : np.array of shape=(n_features,)
                Packet from Reader.

            Returns
            -------
            self : self
                Returns self
            """
        #print(packet)
        try:
            # Extract IPs from packet
            ip_a, ip_b = packet[5], packet[6]
        except BaseException as exp:
            raise ValueError('This is not ip packet.')
        try:
            # Extract ports from packet
            port_a, port_b = int(packet[7]), int(packet[8])
        except BaseException as exp:
            raise ValueError('This ip packet is not sample of tcp or udp or gre.')
        # Perform packet check
        if self.src is not None:
            if {self.src, self.dst} != {ip_a, ip_b} and {self.sport, self.dport} != {port_a, port_b}:
                raise ValueError("Packet {} incompatible with flow {}" .format(packet, self))
        # Set endpoints where smallest dport is destination
        elif port_a > port_b:
            self.src  , self.dst   = ip_a  , ip_b
            self.sport, self.dport = port_a, port_b
        else:
            self.src  , self.dst   = ip_b  , ip_a
            self.sport, self.dport = port_b, port_a

        # Add extension if any
        for i in range(len(packet[-1])):
            #if packet[-1][i] != "":
                if extension[i] not in self.extension:
                    self.extension.setdefault(extension[i],[])
                self.extension[extension[i]].append(packet[-1][i])


        # Set timestamps and lengths
        #print(packet)
        self.ip_timestamps.append(float(packet[3]))
        self.ip_lengths   .append( int(packet[4]) if packet[5] == self.src else
                               -int(packet[4]))
        if int(packet[9]) != 0:
            self.payload_lengths.append( int(packet[9]) if packet[5] == self.src else
                               -int(packet[9]))
            self.payload_timestamps.append(float(packet[3]))

        # Return self
        return self

    ########################################################################
    #                  Source/Destination/Time attributes                  #
    ########################################################################

    @property
    def source(self):
        """(source IP, source port)-tuple of Flow"""
        return (self.src, self.sport)

    @property
    def destination(self):
        """(destination IP, destination port)-tuple of Flow"""
        return (self.dst, self.dport)

    @property
    def time_start(self):
        """Returns start time of Flow"""
        return min(self.timestamps)

    @property
    def time_end(self):
        """Returns end time of Flow"""
        return max(self.timestamps)

    ########################################################################
    #                           Class overrides                            #
    ########################################################################

    def __len__(self):
        """Return length of Flow in packets."""
        return len(self.lengths)

    def __str__(self):
        """Return string representation of flow."""
        if self.main=='ip':
            return "[Time {} to {}] {:>15}:{:<5} <-> {:>15}:{:<5} [IP Packet Size Length {}] [SNI: {}]".format(
                datetime.fromtimestamp(min(self.timestamps)).strftime("%H:%M:%S.%f"),
                datetime.fromtimestamp(max(self.timestamps)).strftime("%H:%M:%S.%f"),
                self.src, self.sport, self.dst, self.dport,
                len(self),self.extension)
        else:
            return "[Time {} to {}] {:>15}:{:<5} <-> {:>15}:{:<5} [Payload Packet Size Length {}] [SNI: {}]".format(
                datetime.fromtimestamp(min(self.timestamps)).strftime("%H:%M:%S.%f"),
                datetime.fromtimestamp(max(self.timestamps)).strftime("%H:%M:%S.%f"),
                self.src, self.sport, self.dst, self.dport,
                len(self),self.extension)
    def __gt__(self, other):
        """Greater than object override"""
        return min(self.timestamps) >  min(other.timestamps)

    def __ge__(self, other):
        """Greater equals object override"""
        return min(self.timestamps) >= min(other.timestamps)

    def __lt__(self, other):
        """Less than object override"""
        return min(self.timestamps) <  min(other.timestamps)

    def __le__(self, other):
        """Less equals object override"""
        return min(self.timestamps) <= min(other.timestamps)